@@ -10,7 +10,7 @@
 # the full text of the license.
 
 import subprocess, tempfile, os.path, re, pwd, grp, os, time, io
-import fnmatch, glob, traceback, errno, sys, atexit, locale, imp, stat
+import fnmatch, glob, traceback, errno, sys, atexit, locale, imp
 
 import xml.dom, xml.dom.minidom
 from xml.parsers.expat import ExpatError
@@ -119,8 +119,15 @@ def _command_output(command, input=None, env=None):
     '''
     sp = subprocess.Popen(command, stdout=subprocess.PIPE,
                           stderr=subprocess.STDOUT, env=env)
-
-    out = sp.communicate(input)[0]
+    # gdb can timeout when trying to retrace some core files, giving it 30
+    # minutes to run should be more than enough.
+    try:
+        out = sp.communicate(input=input, timeout=1800)[0]
+    except subprocess.TimeoutExpired:
+        sp.kill()
+        out, errs = sp.communicate()
+        raise OSError('Error: command %s timedout with exit code %i: %s' % (
+            str(command), sp.returncode, out))
     if sp.returncode == 0:
         return out
     else:
@@ -225,7 +232,7 @@ def _run_hook(report, ui, hook):
                 raise
     except StopIteration:
         return True
-    except Exception as e:
+    except Exception:
         hookname = os.path.splitext(os.path.basename(hook))[0].replace('-', '_')
         report['HookError_' + hookname] = traceback.format_exc()
         apport.error('hook %s crashed:', hook)
@@ -392,8 +399,7 @@ class Report(problem_report.ProblemReport):
         This adds:
         - UserGroups: system groups the user is in
         '''
-        # Use effective uid in case privileges were dropped
-        user = pwd.getpwuid(os.geteuid())[0]
+        user = pwd.getpwuid(os.getuid())[0]
         groups = [name for name, p, gid, memb in grp.getgrall()
                   if user in memb and gid < 1000]
         groups.sort()
@@ -431,7 +437,7 @@ class Report(problem_report.ProblemReport):
             return
 
         cmdargs = self['ProcCmdline'].split('\0')
-        bindirs = ['/bin/', '/sbin/', '/usr/bin/', '/usr/sbin/']
+        bindirs = ['/usr/bin/', '/usr/sbin/', '/bin/', '/sbin/']
 
         # filter out interpreter options
         while len(cmdargs) >= 2 and cmdargs[1].startswith('-'):
@@ -519,7 +525,7 @@ class Report(problem_report.ProblemReport):
             pathlist = [path]
 
             if not module and desc[2] == imp.PKG_DIRECTORY:
-                    module = ['__init__']
+                module = ['__init__']
 
         if path and path.endswith('.pyc'):
             path = path[:-1]
@@ -577,7 +583,9 @@ class Report(problem_report.ProblemReport):
         if 'ExecutablePath' not in self:
             try:
                 self['ExecutablePath'] = _read_proc_link('exe', pid, proc_pid_fd)
-            except OSError as e:
+            except (PermissionError, OSError, FileNotFoundError) as e:
+                if e.errno in (errno.EPERM, errno.EACCES):
+                    raise ValueError('not accessible')
                 if e.errno == errno.ENOENT:
                     raise ValueError('invalid process')
                 else:
@@ -733,7 +741,8 @@ class Report(problem_report.ProblemReport):
         files there.
 
         Raises a IOError if the core dump is invalid/truncated, or OSError if
-        calling gdb fails.
+        calling gdb fails, or FileNotFoundError if gdb or the crashing
+        executable cannot be found.
         '''
         if 'CoreDump' not in self or 'ExecutablePath' not in self:
             return
@@ -746,6 +755,9 @@ class Report(problem_report.ProblemReport):
                        'GLibAssertionMessage': 'print __glib_assert_msg',
                        'NihAssertionMessage': 'print (char*) __nih_abort_msg'}
         gdb_cmd, environ = self.gdb_command(rootdir, gdb_sandbox)
+        if not gdb_cmd:
+            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
+                                    'gdb not found in retracing env')
 
         # limit maximum backtrace depth (to avoid looped stacks)
         gdb_cmd += ['--batch', '--ex', 'set backtrace limit 2000']
@@ -755,6 +767,9 @@ class Report(problem_report.ProblemReport):
         for name, cmd in gdb_reports.items():
             value_keys.append(name)
             gdb_cmd += ['--ex', 'p -99', '--ex', cmd]
+        # End with our separator, ensures gdb's return code is as expected
+        value_keys.append('separator')
+        gdb_cmd += ['--ex', 'p -99']
 
         out = _command_output(gdb_cmd, env=environ).decode('UTF-8', errors='replace')
 
@@ -767,6 +782,9 @@ class Report(problem_report.ProblemReport):
             reason = 'Invalid core dump: ' + warnings.strip()
             self['UnreportableReason'] = reason
             raise IOError(reason)
+        elif out.split('\n')[0].endswith('No such file or directory.'):
+            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
+                                    'executable file for coredump not found')
 
         # split the output into the various fields
         part_re = re.compile(r'^\$\d+\s*=\s*-99$', re.MULTILINE)
@@ -988,37 +1006,31 @@ class Report(problem_report.ProblemReport):
 
         Raises ValueError if the file exists but is invalid XML.
         '''
-        # Properly handle dropped privileges
-        homedir = pwd.getpwuid(os.geteuid())[5]
-        ifpath = _ignore_file.replace("~", homedir)
-
-        contents = ''
-        fd = None
-        f = None
-
+        orig_home = os.getenv('HOME')
+        if orig_home is not None:
+            del os.environ['HOME']
+        ifpath = os.path.expanduser(_ignore_file)
+        if orig_home is not None:
+            os.environ['HOME'] = orig_home
+        euid = os.geteuid()
         try:
-            fd = os.open(ifpath, os.O_NOFOLLOW | os.O_RDONLY)
-            st = os.fstat(fd)
-            if stat.S_ISREG(st.st_mode):
-                f = os.fdopen(fd, 'r')
-                # Limit size to prevent DoS
-                contents = f.read(50000)
+            # drop permissions temporarily to try open users ignore file
+            os.seteuid(os.getuid())
+            fp = open(ifpath, 'r')
         except (IOError, OSError):
-            pass
+            fp = None
         finally:
-            if f is not None:
-                f.close()
-            elif fd is not None:
-                os.close(fd)
-
-        if contents == '':
+            os.seteuid(euid)
+        if fp is None or os.fstat(fp.fileno()).st_size == 0:
             # create a document from scratch
             dom = xml.dom.getDOMImplementation().createDocument(None, 'apport', None)
         else:
             try:
-                dom = xml.dom.minidom.parseString(contents)
+                dom = xml.dom.minidom.parse(fp)
             except ExpatError as e:
                 raise ValueError('%s has invalid format: %s' % (_ignore_file, str(e)))
+        if fp is not None:
+            fp.close()
         # remove whitespace so that writing back the XML does not accumulate
         # whitespace
         dom.documentElement.normalize()
@@ -1038,8 +1050,6 @@ class Report(problem_report.ProblemReport):
 
         This requires the ExecutablePath attribute. Throws a ValueError if the
         file has an invalid format.
-
-        Privileges may need to be dropped before calling this.
         '''
         assert 'ExecutablePath' in self
 
@@ -1086,13 +1096,10 @@ class Report(problem_report.ProblemReport):
             return False
 
         # search for existing entry and update it
-        try:
-            for ignore in dom.getElementsByTagName('ignore'):
-                if ignore.getAttribute('program') == self['ExecutablePath']:
-                    if float(ignore.getAttribute('mtime')) >= cur_mtime:
-                        return True
-        except (ValueError, KeyError):
-            pass
+        for ignore in dom.getElementsByTagName('ignore'):
+            if ignore.getAttribute('program') == self['ExecutablePath']:
+                if float(ignore.getAttribute('mtime')) >= cur_mtime:
+                    return True
 
         return False
 
@@ -1105,8 +1112,6 @@ class Report(problem_report.ProblemReport):
 
         Throws a ValueError if the file already exists and has an invalid
         format.
-
-        Privileges may need to be dropped before calling this.
         '''
         assert 'ExecutablePath' in self
 
@@ -1132,10 +1137,14 @@ class Report(problem_report.ProblemReport):
             e.setAttribute('mtime', mtime)
             dom.documentElement.appendChild(e)
 
-        # Write back file
-        # Properly handle dropped privileges
-        homedir = pwd.getpwuid(os.geteuid())[5]
-        ignore_file_path = _ignore_file.replace("~", homedir)
+        # write back file; temporarily unset $HOME, as this gets the wrong home
+        # dir for e. g. sudo
+        orig_home = os.getenv('HOME')
+        if orig_home is not None:
+            del os.environ['HOME']
+        ignore_file_path = os.path.expanduser(_ignore_file)
+        if orig_home is not None:
+            os.environ['HOME'] = orig_home
 
         with open(ignore_file_path, 'w') as fd:
             dom.writexml(fd, addindent='  ', newl='\n')
@@ -1530,10 +1539,9 @@ class Report(problem_report.ProblemReport):
         removes the ProcCwd attribute completely.
         '''
         replacements = []
-        # Do not replace "root"
         if (os.getuid() > 0):
-            # Use effective uid in case privileges were dropped
-            p = pwd.getpwuid(os.geteuid())
+            # do not replace "root"
+            p = pwd.getpwuid(os.getuid())
             if len(p[0]) >= 2:
                 replacements.append((re.compile(r'\b%s\b' % re.escape(p[0])), 'username'))
             replacements.append((re.compile(r'\b%s\b' % re.escape(p[5])), '/home/username'))
@@ -1589,8 +1597,7 @@ class Report(problem_report.ProblemReport):
                            if gdb_sandbox else None)
         gdb_path = _which_extrapath('gdb', gdb_sandbox_bin)
         if not gdb_path:
-            apport.fatal('gdb does not exist in the %ssandbox nor on the host'
-                         % ('gdb ' if not same_arch else ''))
+            return '', ''
         command = [gdb_path]
         environ = None
 
@@ -1614,8 +1621,8 @@ class Report(problem_report.ProblemReport):
                         '--ex', 'add-auto-load-safe-path ' + sandbox,
                         # needed to fix /lib64/ld-linux-x86-64.so.2 broken
                         # symlink
-                        '--ex', 'set solib-search-path %s/lib/%s' % \
-                        (sandbox, native_multiarch)]
+                        '--ex', 'set solib-search-path %s/lib/%s:%s/usr/lib/%s' % \
+                        (sandbox, native_multiarch, sandbox, native_multiarch)]
             if gdb_sandbox:
                 ld_lib_path = '%s/lib:%s/lib/%s:%s/usr/lib/%s:%s/usr/lib' % \
                     (gdb_sandbox, gdb_sandbox, native_multiarch,
@@ -1632,7 +1639,6 @@ class Report(problem_report.ProblemReport):
                             gdb_sandbox]
             executable = sandbox + '/' + executable
 
-        assert os.path.exists(executable)
         command += ['--ex', 'file "%s"' % executable]
 
         if 'CoreDump' in self:
