@@ -26,6 +26,7 @@
 from __future__ import absolute_import, print_function
 
 import apt_pkg
+import apt
 import copy
 from hashlib import md5
 import re
@@ -137,8 +138,7 @@ class SoftwareProperties(object):
     " wait for all running threads (PPA key fetchers) to exit "
     for t in threading.enumerate():
       if t.ident != threading.current_thread().ident:
-        if not t.daemon:
-            t.join()
+        t.join()
 
   def backup_apt_conf(self):
     """Backup all apt configuration options"""
@@ -695,8 +695,7 @@ class SoftwareProperties(object):
       keyp = "%s/%s.key" % (self.CHANNEL_PATH, channel)
       self.add_key(keyp)
 
-    cdata = (shortcut.add_key, {'keyserver': (self.options and
-                                              self.options.keyserver)})
+    cdata = (shortcut.add_key, {'keyserver': (self.options)})
     def addkey_func():
         func, kwargs = cdata
         msg = "Added key."
@@ -834,7 +833,6 @@ class SoftwareProperties(object):
         return False
     try:
         res = self.apt_key.add(path)
-        self.KeysModified()
         return res
     except:
         return False
@@ -850,7 +848,6 @@ class SoftwareProperties(object):
     """Remove a gnupg key from the list of trusted software vendors"""
     try:
         self.apt_key.rm(keyid)
-        self.KeysModified()
         return True
     except:
         return False
@@ -859,11 +856,16 @@ class SoftwareProperties(object):
     """ Run apt-key update """
     try:
       self.apt_key.update()
-      self.KeysModified()
       return True
     except:
       return False
 
+  def get_package_id(self, ver):
+    """ Return the PackageKit package id """
+    assert isinstance(ver, apt.package.Version)
+    return "%s;%s;%s;" % (ver.package.shortname, ver.version,
+                          ver.package.architecture())
+
 def shortcut_handler(shortcut):
     for factory in _SHORTCUT_FACTORIES:
         ret = factory(shortcut)
