@@ -1,6 +1,6 @@
 '''common.py: common classes for ufw'''
 #
-# Copyright 2008-2013 Canonical Ltd.
+# Copyright 2008-2018 Canonical Ltd.
 #
 #    This program is free software: you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License version 3,
@@ -29,6 +29,7 @@ prefix_dir = "/usr"
 iptables_dir = "/sbin"
 do_checks = True
 
+
 class UFWError(Exception):
     '''This class represents ufw exceptions'''
     def __init__(self, value):
@@ -41,7 +42,8 @@ class UFWError(Exception):
 class UFWRule:
     '''This class represents firewall rules'''
     def __init__(self, action, protocol, dport="any", dst="0.0.0.0/0",
-                 sport="any", src="0.0.0.0/0", direction="in", forward=False):
+                 sport="any", src="0.0.0.0/0", direction="in", forward=False,
+                 comment=""):
         # Be sure to update dup_rule accordingly...
         self.remove = False
         self.updated = False
@@ -61,6 +63,7 @@ class UFWRule:
         self.interface_out = ""
         self.direction = ""
         self.forward = forward
+        self.comment = ""
         try:
             self.set_action(action)
             self.set_protocol(protocol)
@@ -69,6 +72,7 @@ class UFWRule:
             self.set_src(src)
             self.set_dst(dst)
             self.set_direction(direction)
+            self.set_comment(comment)
         except UFWError:
             raise
 
@@ -102,7 +106,8 @@ class UFWRule:
         rule.interface_in = self.interface_in
         rule.interface_out = self.interface_out
         rule.direction = self.direction
-        rule.forward = self.forward 
+        rule.forward = self.forward
+        rule.comment = self.comment
 
         return rule
 
@@ -241,12 +246,7 @@ class UFWRule:
 
     def set_protocol(self, protocol):
         '''Sets protocol of the rule'''
-        if protocol == "tcp" or \
-           protocol == "udp" or \
-           protocol == "ipv6" or \
-           protocol == "esp" or \
-           protocol == "ah" or \
-           protocol == "any":
+        if protocol in ufw.util.supported_protocols + ['any']:
             self.protocol = protocol
         else:
             err_msg = _("Unsupported protocol '%s'") % (protocol)
@@ -294,22 +294,47 @@ class UFWRule:
 
     def set_interface(self, if_type, name):
         '''Sets an interface for rule'''
+        # libxtables/xtables.c xtables_parse_interface() specifies
+        # - < 16
+        # - not empty
+        # - doesn't contain ' '
+        # - doesn't contain '/'
+        #
+        # net/core/dev.c from the kernel specifies:
+        # - < 16
+        # - not empty
+        # - != '.' or '..'
+        # - doesn't contain '/', ':' or whitespace
         if if_type != "in" and if_type != "out":
             err_msg = _("Bad interface type")
             raise UFWError(err_msg)
 
+        # Separate a few of the invalid checks out so we can give a nice error
         if '!' in str(name):
             err_msg = _("Bad interface name: reserved character: '!'")
             raise UFWError(err_msg)
 
-        if not re.match(r'^[a-zA-Z][a-zA-Z0-9:]*[a-zA-Z0-9]', str(name)):
-            err_msg = _("Bad interface name")
-            raise UFWError(err_msg)
-
         if ':' in str(name):
             err_msg = _("Bad interface name: can't use interface aliases")
             raise UFWError(err_msg)
 
+        if str(name) == "." or str(name) == "..":
+            err_msg = _("Bad interface name: can't use '.' or '..'")
+            raise UFWError(err_msg)
+
+        if (len(str(name)) == 0):
+            err_msg = _("Bad interface name: interface name is empty")
+            raise UFWError(err_msg)
+
+        if (len(str(name)) > 15):
+            err_msg = _("Bad interface name: interface name too long")
+            raise UFWError(err_msg)
+
+        # We are going to limit this even further to avoid shell meta
+        if not re.match(r'^[a-zA-Z0-9_\-\.\+,=%@]+$', str(name)):
+            err_msg = _("Bad interface name")
+            raise UFWError(err_msg)
+
         if if_type == "in":
             self.interface_in = name
         else:
@@ -317,7 +342,10 @@ class UFWRule:
 
     def set_position(self, num):
         '''Sets the position of the rule'''
-        if not re.match(r'^[0-9]+', str(num)):
+        # -1 prepend
+        #  0 append
+        # >0 insert
+        if str(num) != "-1" and not re.match(r'^[0-9]+', str(num)):
             err_msg = _("Insert position '%s' is not a valid position") % (num)
             raise UFWError(err_msg)
         self.position = int(num)
@@ -339,6 +367,14 @@ class UFWRule:
             err_msg = _("Unsupported direction '%s'") % (direction)
             raise UFWError(err_msg)
 
+    def get_comment(self):
+        '''Get decoded comment of the rule'''
+        return ufw.util.hex_decode(self.comment)
+
+    def set_comment(self, comment):
+        '''Sets comment of the rule'''
+        self.comment = comment
+
     def normalize(self):
         '''Normalize src and dst to standard form'''
         changed = False
@@ -379,7 +415,8 @@ class UFWRule:
         Return codes:
           0  match
           1  no match
-         -1  match all but action
+         -1  match all but action, log-type and/or comment
+         -2  match all but comment
         '''
         if not x or not y:
             raise ValueError()
@@ -421,21 +458,28 @@ class UFWRule:
         if x.forward != y.forward:
             debug(dbg_msg)
             return 1
-        if x.action == y.action and x.logtype == y.logtype:
+        if x.action == y.action and x.logtype == y.logtype and \
+                x.comment == y.comment:
             dbg_msg = _("Found exact match")
             debug(dbg_msg)
             return 0
+        if x.action == y.action and x.logtype == y.logtype and \
+                x.comment != y.comment:
+            dbg_msg = _("Found exact match, excepting comment")
+            debug(dbg_msg)
+            return -2
 
-        dbg_msg = _("Found non-action/non-logtype match " \
-                    "(%(xa)s/%(ya)s %(xl)s/%(yl)s)") % \
-                    ({'xa': x.action, 'ya': y.action, \
-                      'xl': x.logtype, 'yl': y.logtype})
+        dbg_msg = _("Found non-action/non-logtype/comment match " \
+                    "(%(xa)s/%(ya)s/'%(xc)s' %(xl)s/%(yl)s/'%(yc)s')") % \
+                    ({'xa': x.action, 'ya': y.action,
+                      'xl': x.logtype, 'yl': y.logtype,
+                      'xc': x.comment, 'yc': y.comment})
         debug(dbg_msg)
         return -1
 
     def fuzzy_dst_match(x, y):
         '''This will match if x is more specific than y. Eg, for protocol if x
-	   is tcp and y is all or for address if y is a network and x is a
+           is tcp and y is all or for address if y is a network and x is a
            subset of y (where x is either an address or network). Returns:
 
             0  match
@@ -493,11 +537,11 @@ class UFWRule:
             return 1
 
         if y.interface_in == "":
-	    # If destination interface is not specified, destination addresses
+            # If destination interface is not specified, destination addresses
             # must match or x must be contained in y
 
             if x.interface_in == "" and x._is_anywhere(x.dst):
-		# if x and y interfaces are not specified, and x.dst is
+                # if x and y interfaces are not specified, and x.dst is
                 # anywhere then ok
                 pass
             elif x.dst != y.dst and '/' not in y.dst:
@@ -509,10 +553,10 @@ class UFWRule:
                       (x.dst, y.dst))
                 return 1
         else:
-	    # If destination interface is specified, then:
+            # If destination interface is specified, then:
             #  if specified, both interfaces must match or
             #  the IP of the interface must match the IP of y or
-	    #  the IP of the interface must be contained in y
+            #  the IP of the interface must be contained in y
             if x.interface_in != "" and x.interface_in != y.interface_in:
                 debug("(interface) " + dbg_msg + " (%s != %s)" % \
                       (x.interface_in, y.interface_in))
@@ -551,14 +595,15 @@ class UFWRule:
 
     def get_app_tuple(self):
         '''Returns a tuple to identify an app rule. Tuple is:
-             dapp dst sapp src
+             dapp dst sapp src direction_iface|direction
            or
-             dport dst sapp src
+             dport dst sapp src direction_iface|direction
            or
-             dapp dst sport src
+             dapp dst sport src direction_iface|direction
 
-           All of these might have in_eth0 out_eth0 (or similar) if an
-           interface is also defined.
+           where direction_iface is of form 'in_eth0', 'out_eth0' or
+           'in_eth0 out_eth0' (ie, both interfaces used). If no interfaces are
+           specified, then tuple ends with the direction instead.
         '''
         tupl = ""
         if self.dapp != "" or self.sapp != "":
@@ -570,11 +615,36 @@ class UFWRule:
                 tupl = "%s %s %s %s" % (self.dapp, self.dst, self.sport, \
                                          self.src)
 
-            # add interfaces to the end, if they exist
-            if self.interface_in != "":
-                tupl += " in_%s" % (self.interface_in)
-            if self.interface_out != "":
-                tupl += " out_%s" % (self.interface_out)
+            # if neither interface exists, add the direction
+            if self.interface_in == "" and self.interface_out == "":
+                tupl += " %s" % (self.direction)
+            # otherwise, add the interfaces
+            else:
+                if self.interface_in != "":
+                    tupl += " in_%s" % (self.interface_in)
+                if self.interface_out != "":
+                    tupl += " out_%s" % (self.interface_out)
 
         return tupl
 
+    def verify(self, rule_iptype):
+        '''Verify rule'''
+        # Verify protocol not specified with application rule
+        if self.protocol != "any" and \
+           (self.sapp != "" or self.dapp != ""):
+            err_msg = _("Improper rule syntax ('%s' specified with app rule)") \
+                        % (self.protocol)
+            raise UFWError(err_msg)
+
+        if self.protocol in ufw.util.ipv4_only_protocols and \
+           rule_iptype == "v6":
+            # Can't use protocol these protocols with v6 addresses
+            err_msg = _("Invalid IPv6 address with protocol '%s'") % \
+                        (self.protocol)
+            raise UFWError(err_msg)
+
+        if self.protocol in ufw.util.portless_protocols:
+            if self.dport != "any" or self.sport != "any":
+                err_msg = _("Invalid port with protocol '%s'") % \
+                            (self.protocol)
+                raise UFWError(err_msg)
